@@ -26,6 +26,11 @@
 
 #include <gio/gio.h>
 
+#ifdef BACKEND_IS_GNUTLS
+#include <gnutls/gnutls.h>
+#include <gnutls/pkcs11.h>
+#endif
+
 #include <sys/types.h>
 #include <string.h>
 
@@ -243,6 +248,27 @@ test_create_certificate_with_garbage_input (TestCertificate *test,
   g_clear_error (&error);
 }
 
+static void
+test_create_certificate_pkcs11 (TestCertificate *test,
+                                gconstpointer data)
+{
+#if !GLIB_CHECK_VERSION (2, 63, 0)
+  g_test_skip ("glib-networking built against too old GLib");
+#elif !defined (BACKEND_IS_GNUTLS)
+  g_test_skip ("This backend does not support PKCS #11");
+#else
+  GTlsCertificate *cert;
+  GError *error = NULL;
+
+  cert = g_initable_new (test->cert_gtype, NULL, &error,
+                         "pkcs11-uri", "pkcs11:model=mock;token=Mock%20Certificate;object=Mock%20Certificate",
+                         NULL);
+
+  g_assert_no_error (error);
+  g_assert_nonnull (cert);
+#endif
+}
+
 static void
 test_create_certificate_chain (void)
 {
@@ -556,44 +582,57 @@ int
 main (int   argc,
       char *argv[])
 {
+#ifdef BACKEND_IS_GNUTLS
+  char *module_path;
+#endif
+
   g_test_init (&argc, &argv, NULL);
 
   g_setenv ("GSETTINGS_BACKEND", "memory", TRUE);
   g_setenv ("GIO_USE_TLS", BACKEND, TRUE);
   g_assert (g_ascii_strcasecmp (G_OBJECT_TYPE_NAME (g_tls_backend_get_default ()), "GTlsBackend" BACKEND) == 0);
 
-  g_test_add ("/tls/certificate/create-pem", TestCertificate, NULL,
+#ifdef BACKEND_IS_GNUTLS
+  module_path = g_build_filename (g_getenv ("G_TEST_BUILDDIR"), "mock-pkcs11.so", NULL);
+  g_assert (gnutls_pkcs11_init (GNUTLS_PKCS11_FLAG_MANUAL, NULL) == GNUTLS_E_SUCCESS);
+  g_assert (gnutls_pkcs11_add_provider (module_path, NULL) == GNUTLS_E_SUCCESS);
+  g_free (module_path);
+#endif
+
+  g_test_add ("/tls/" BACKEND "/certificate/create-pem", TestCertificate, NULL,
               setup_certificate, test_create_pem, teardown_certificate);
-  g_test_add ("/tls/certificate/create-der", TestCertificate, NULL,
+  g_test_add ("/tls/" BACKEND "/certificate/create-der", TestCertificate, NULL,
               setup_certificate, test_create_der, teardown_certificate);
-  g_test_add ("/tls/certificate/create-with-key-pem", TestCertificate, NULL,
+  g_test_add ("/tls/" BACKEND "/certificate/create-with-key-pem", TestCertificate, NULL,
               setup_certificate, test_create_with_key_pem, teardown_certificate);
-  g_test_add ("/tls/certificate/create-with-key-der", TestCertificate, NULL,
+  g_test_add ("/tls/" BACKEND "/certificate/create-with-key-der", TestCertificate, NULL,
               setup_certificate, test_create_with_key_der, teardown_certificate);
-  g_test_add ("/tls/certificate/create-with-issuer", TestCertificate, NULL,
+  g_test_add ("/tls/" BACKEND "/certificate/create-with-issuer", TestCertificate, NULL,
               setup_certificate, test_create_certificate_with_issuer, teardown_certificate);
-  g_test_add ("/tls/certificate/create-with-garbage-input", TestCertificate, NULL,
+  g_test_add ("/tls/" BACKEND "/certificate/create-with-garbage-input", TestCertificate, NULL,
               setup_certificate, test_create_certificate_with_garbage_input, teardown_certificate);
+  g_test_add ("/tls/" BACKEND "/certificate/pkcs11", TestCertificate, NULL,
+              setup_certificate, test_create_certificate_pkcs11, teardown_certificate);
 
-  g_test_add_func ("/tls/certificate/create-chain", test_create_certificate_chain);
-  g_test_add_func ("/tls/certificate/create-no-chain", test_create_certificate_no_chain);
-  g_test_add_func ("/tls/certificate/create-list", test_create_list);
-  g_test_add_func ("/tls/certificate/create-list-bad", test_create_list_bad);
+  g_test_add_func ("/tls/" BACKEND "/certificate/create-chain", test_create_certificate_chain);
+  g_test_add_func ("/tls/" BACKEND "/certificate/create-no-chain", test_create_certificate_no_chain);
+  g_test_add_func ("/tls/" BACKEND "/certificate/create-list", test_create_list);
+  g_test_add_func ("/tls/" BACKEND "/certificate/create-list-bad", test_create_list_bad);
 
-  g_test_add ("/tls/certificate/verify-good", TestVerify, NULL,
+  g_test_add ("/tls/" BACKEND "/certificate/verify-good", TestVerify, NULL,
               setup_verify, test_verify_certificate_good, teardown_verify);
-  g_test_add ("/tls/certificate/verify-bad-identity", TestVerify, NULL,
+  g_test_add ("/tls/" BACKEND "/certificate/verify-bad-identity", TestVerify, NULL,
               setup_verify, test_verify_certificate_bad_identity, teardown_verify);
-  g_test_add ("/tls/certificate/verify-bad-ca", TestVerify, NULL,
+  g_test_add ("/tls/" BACKEND "/certificate/verify-bad-ca", TestVerify, NULL,
               setup_verify, test_verify_certificate_bad_ca, teardown_verify);
-  g_test_add ("/tls/certificate/verify-bad-before", TestVerify, NULL,
+  g_test_add ("/tls/" BACKEND "/certificate/verify-bad-before", TestVerify, NULL,
               setup_verify, test_verify_certificate_bad_before, teardown_verify);
-  g_test_add ("/tls/certificate/verify-bad-expired", TestVerify, NULL,
+  g_test_add ("/tls/" BACKEND "/certificate/verify-bad-expired", TestVerify, NULL,
               setup_verify, test_verify_certificate_bad_expired, teardown_verify);
-  g_test_add ("/tls/certificate/verify-bad-combo", TestVerify, NULL,
+  g_test_add ("/tls/" BACKEND "/certificate/verify-bad-combo", TestVerify, NULL,
               setup_verify, test_verify_certificate_bad_combo, teardown_verify);
 
-  g_test_add_func ("/tls/certificate/is-same", test_certificate_is_same);
+  g_test_add_func ("/tls/" BACKEND "/certificate/is-same", test_certificate_is_same);
 
   return g_test_run();
 }
